@@ -116,11 +116,13 @@ struct write_window {
 };
 
 static pthread_mutex_t window_lock = PTHREAD_MUTEX_INITIALIZER;
+static pthread_mutex_t window_fd_lock = PTHREAD_MUTEX_INITIALIZER;
 static struct write_window window[NR_WINDOWS];
 
 static int
 evict_writes (int fd, uint64_t offset, size_t len)
 {
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&window_fd_lock);
   struct write_window oldest = { 0 };
 
   {
@@ -146,10 +148,14 @@ evict_writes (int fd, uint64_t offset, size_t len)
     window[NR_WINDOWS-1].offset = offset;
     window[NR_WINDOWS-1].len = len;
   }
-  /* Release lock here. */
+  /* Release window_lock here, but hold window_fd_lock since we don't
+   * want file_close -> remove_fd_from_window to close oldest.fd
+   */
 
   /* Evict the oldest window from the page cache (synchronously). */
   if (oldest.len > 0) {
+    int r;
+
     if (sync_file_range (oldest.fd, oldest.offset, oldest.len,
                          SYNC_FILE_RANGE_WAIT_BEFORE |
                          SYNC_FILE_RANGE_WRITE |
@@ -157,9 +163,12 @@ evict_writes (int fd, uint64_t offset, size_t len)
       nbdkit_error ("sync_file_range: cache=none: evicting oldest window: %m");
       return -1;
     }
-    if (posix_fadvise (oldest.fd, oldest.offset, oldest.len,
-                       POSIX_FADV_DONTNEED) == -1)
+    r = posix_fadvise (oldest.fd, oldest.offset, oldest.len,
+                       POSIX_FADV_DONTNEED);
+    if (r != 0) {
+      errno = r;
       nbdkit_debug ("posix_fadvise: POSIX_FADV_DONTNEED: (ignored): %m");
+    }
   }
 
   return 0;
@@ -171,7 +180,7 @@ evict_writes (int fd, uint64_t offset, size_t len)
 static void
 remove_fd_from_window (int fd)
 {
-  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&window_lock);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&window_fd_lock);
   size_t i;
 
   for (i = 0; i < NR_WINDOWS; ++i)
@@ -677,8 +686,10 @@ file_open (int readonly)
     /* This is a hint so we ignore failures. */
 #ifdef HAVE_POSIX_FADVISE
     int r = posix_fadvise (h->fd, 0, 0, fadvise_mode);
-    if (r == -1)
-      nbdkit_debug ("posix_fadvise: %s: %m (ignored)", file);
+    if (r != 0) {
+      errno = r;
+      nbdkit_debug ("posix_fadvise: %s: (ignored): %m", file);
+    }
 #else
     nbdkit_debug ("fadvise is not supported");
 #endif
